import numpy as np
import utool as ut  # NOQA
import ubelt as ub  # NOQA
import cv2


class IntensityPreproc(object):
    """
    Prefered over old methods

    CommandLine:
        python -m vtool_ibeis.image_filters IntensityPreproc --show

    Doctest:
        >>> from vtool_ibeis.image_filters import *
        >>> import vtool_ibeis as vt
        >>> chipBGR = vt.imread(ut.grab_file_url('http://i.imgur.com/qVWQaex.jpg'))
        >>> filter_list = [
        >>>     ('medianblur', {}),
        >>>     ('adapteq', {}),
        >>> ]
        >>> self = IntensityPreproc()
        >>> chipBGR2 = self.preprocess(chipBGR, filter_list)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(chipBGR, pnum=(1, 2, 1), fnum=1)
        >>> pt.imshow(chipBGR2, pnum=(1, 2, 2), fnum=1)
        >>> ut.show_if_requested()
    """

    def preprocess(self, chipBGR, filter_list):
        """
        filter_list is a list of (name, config) tuples for preforming filter ops
        """

        # Convert into LAB space for grayscale extraction
        chipLAB = cv2.cvtColor(chipBGR, cv2.COLOR_BGR2LAB)
        intensity = chipLAB[:, :, 0]

        # Modify intensity
        for filtname, config in filter_list:
            intensity = getattr(self, filtname)(intensity, **config)

        # Add color back in
        chipLAB[:, :, 0] = intensity
        chipBGR = cv2.cvtColor(chipLAB, cv2.COLOR_LAB2BGR)
        return chipBGR

    def adapteq(self, intensity, tileGridSize=(8, 8), clipLimit=2.0):
        clahe_obj = cv2.createCLAHE(clipLimit, tileGridSize)
        intensity = clahe_obj.apply(intensity)
        return intensity

    def medianblur(self, intensity, noise_thresh=50, ksize1=3, ksize2=5):
        istd = intensity.std()
        ksize = ksize1 if istd < noise_thresh else ksize2
        intensity = cv2.medianBlur(intensity, ksize)
        return intensity

    def histeq(self, intensity):
        """ Histogram equalization of a grayscale image. """
        return cv2.equalizeHist(intensity)


def manta_matcher_filters(chipBGR):
    """
    References:
        http://onlinelibrary.wiley.com/doi/10.1002/ece3.587/full

    Ignore:
        >>> from ibeis.core_annots import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb('Mantas')
        >>> chipBGR = vt.imread(ut.grab_file_url('http://i.imgur.com/qVWQaex.jpg'))
    """
    chipLAB = cv2.cvtColor(chipBGR, cv2.COLOR_BGR2LAB)

    intensity = chipLAB[:, :, 0]
    # Median filter
    noise_thresh = 100
    ksize = 5 if intensity.std() > noise_thresh else 3
    intensity = cv2.medianBlur(intensity, ksize)

    tileGridSize = (8, 8)
    clipLimit = 2.0
    clahe_obj = cv2.createCLAHE(clipLimit, tileGridSize)
    intensity = clahe_obj.apply(intensity, dst=intensity)

    chipLAB[:, :, 0] = intensity
    chipBGR = cv2.cvtColor(chipLAB, cv2.COLOR_LAB2BGR)
    return chipBGR


def adapteq_fn(chipBGR):
    """
    adaptive histogram equalization with CLAHE

    Example:
        >>> from vtool_ibeis.image_filters import *
        >>> import vtool_ibeis as vt
        >>> import utool as ut
        >>> chipBGR = vt.imread(ut.grab_file_url('http://i.imgur.com/qVWQaex.jpg'))
        >>> chip2 = adapteq_fn(chipBGR)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(chipBGR, pnum=(1, 2, 1), fnum=1)
        >>> pt.imshow(chip2, pnum=(1, 2, 2), fnum=1)
        >>> ut.show_if_requested()
    """
    chipLAB = cv2.cvtColor(chipBGR, cv2.COLOR_BGR2LAB)
    tileGridSize = (8, 8)
    clipLimit = 2.0
    clahe_obj = cv2.createCLAHE(clipLimit, tileGridSize)
    chipLAB[:, :, 0] = clahe_obj.apply(chipLAB[:, :, 0])
    chipBGR = cv2.cvtColor(chipLAB, cv2.COLOR_LAB2BGR)
    return chipBGR


def medianfilter_fn(chipBGR):
    """
    median filtering

    Example:
        >>> from vtool_ibeis.image_filters import *
        >>> import vtool_ibeis as vt
        >>> import utool as ut
        >>> chipBGR = vt.imread(ut.grab_file_url('http://i.imgur.com/qVWQaex.jpg'))
        >>> chip2 = adapteq_fn(chipBGR)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(chipBGR, pnum=(1, 2, 1), fnum=1)
        >>> pt.imshow(chip2, pnum=(1, 2, 2), fnum=1)
        >>> ut.show_if_requested()
    """
    chipLAB = cv2.cvtColor(chipBGR, cv2.COLOR_BGR2LAB)
    intensity = chipLAB[:, :, 0]
    noise_thresh = 100
    ksize = 5 if intensity.std() > noise_thresh else 3
    intensity = cv2.medianBlur(intensity, ksize)
    chipLAB[:, :, 0] = intensity
    chipBGR = cv2.cvtColor(chipLAB, cv2.COLOR_LAB2BGR)
    return chipBGR


def histeq_fn(chipBGR):
    """ Histogram equalization of a grayscale image. """
    chipLAB = cv2.cvtColor(chipBGR, cv2.COLOR_BGR2LAB)
    chipLAB[:, :, 0] = cv2.equalizeHist(chipLAB[:, :, 0])
    chipBGR = cv2.cvtColor(chipLAB, cv2.COLOR_LAB2BGR)
    return chipBGR


def clean_mask(mask, num_dilate=3, num_erode=3, window_frac=.025):
    """ Clean the mask
    (num_erode, num_dilate) = (1, 1)
    (w, h) = (10, 10) """
    w = h = int(round(min(mask.shape) * window_frac))
    element = cv2.getStructuringElement(cv2.MORPH_CROSS, (w, h))
    _mask = mask
    # compute the closing
    for ix in range(num_dilate):
        _mask = cv2.dilate(_mask, element)
    for ix in range(num_erode):
        _mask = cv2.erode(_mask, element)
    return _mask


def grabcut_fn(chipBGR):
    """ naively segments a chip """
    chipRGB = cv2.cvtColor(chipBGR, cv2.COLOR_BGR2RGB)
    (h, w) = chipRGB.shape[0:2]
    _mask = np.zeros((h, w), dtype=np.uint8)  # Initialize: mask
    # Set inside to cv2.GC_PR_FGD (probably forground)
    _mask[ :, :] = cv2.GC_PR_FGD
    # Set border to cv2.GC_BGD (definitely background)
    _mask[ 0, :] = cv2.GC_BGD
    _mask[-1, :] = cv2.GC_BGD
    _mask[:,  0] = cv2.GC_BGD
    _mask[:, -1] = cv2.GC_BGD
    # Grab Cut Parameters
    rect = (0, 0, w, h)
    num_iters = 5
    mode = cv2.GC_INIT_WITH_MASK
    bgd_model = np.zeros((1, 13 * 5), np.float64)
    fgd_model = np.zeros((1, 13 * 5), np.float64)
    # Grab Cut Execution
    cv2.grabCut(chipRGB, _mask, rect, bgd_model, fgd_model, num_iters, mode=mode)
    is_forground = (_mask == cv2.GC_FGD) + (_mask == cv2.GC_PR_FGD)
    chip_mask = np.where(is_forground, 255, 0).astype('uint8')
    # Crop
    chip_mask = clean_mask(chip_mask)
    chip_mask = np.array(chip_mask, float) / 255.0
    # Mask value component of HSV space
    chipHSV = cv2.cvtColor(chipRGB, cv2.COLOR_RGB2HSV)
    chipHSV = np.array(chipHSV, dtype=float) / 255.0
    chipHSV[:, :, 2] *= chip_mask
    chipHSV = np.array(np.round(chipHSV * 255.0), dtype=np.uint8)
    seg_chipBGR = cv2.cvtColor(chipHSV, cv2.COLOR_HSV2BGR)
    return seg_chipBGR



if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.image_filters
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
References:
    http://www.exiv2.org/tags.html

TODO:
    https://github.com/recurser/exif-orientation-examples
"""
from PIL.ExifTags import TAGS, GPSTAGS
import PIL.ExifTags  # NOQA
from PIL import Image
import utool as ut
import warnings
from utool import util_time
from vtool_ibeis import image_shared


# Inverse of PIL.ExifTags.TAGS
EXIF_TAG_TO_TAGID = {val: key for (key, val) in TAGS.items()}
GPS_TAG_TO_GPSID  = {val: key for (key, val) in GPSTAGS.items()}

# Relevant EXIF Tags
#'GPSInfo': 34853
#'SensitivityType': 34864  # UNSUPPORTED

GPSINFO_CODE = EXIF_TAG_TO_TAGID['GPSInfo']
DATETIMEORIGINAL_TAGID = EXIF_TAG_TO_TAGID['DateTimeOriginal']
SENSITIVITYTYPE_CODE = 34864  # UNSUPPORTED BY PIL

ORIENTATION_CODE = EXIF_TAG_TO_TAGID['Orientation']
ORIENTATION_UNDEFINED = 'UNDEFINED'
ORIENTATION_000 = 'Normal'
ORIENTATION_090 = '90 Clockwise'
ORIENTATION_180 = 'Upside-Down'
ORIENTATION_270 = '90 Counter-Clockwise'

ORIENTATION_DICT = {
    0: ORIENTATION_UNDEFINED,
    1: ORIENTATION_000,
    2: None,  # Flip Left-to-Right
    3: ORIENTATION_180,
    4: None,  # Flip Top-to-Bottom
    5: None,  # Flip Left-to-Right then Rotate 90
    6: ORIENTATION_090,
    7: None,  # Flip Left-to-Right then Rotate 270
    8: ORIENTATION_270,
}

ORIENTATION_DICT_INVERSE = {
    ORIENTATION_UNDEFINED : 0,
    ORIENTATION_000       : 1,
    ORIENTATION_180       : 3,
    ORIENTATION_090       : 6,
    ORIENTATION_270       : 8,
}

ORIENTATION_ORDER_LIST = [
    ORIENTATION_DICT_INVERSE[ORIENTATION_000],
    ORIENTATION_DICT_INVERSE[ORIENTATION_090],
    ORIENTATION_DICT_INVERSE[ORIENTATION_180],
    ORIENTATION_DICT_INVERSE[ORIENTATION_270],
]


GPSLATITUDE_CODE     = GPS_TAG_TO_GPSID['GPSLatitude']
GPSLATITUDEREF_CODE  = GPS_TAG_TO_GPSID['GPSLatitudeRef']
GPSLONGITUDE_CODE    = GPS_TAG_TO_GPSID['GPSLongitude']
GPSLONGITUDEREF_CODE = GPS_TAG_TO_GPSID['GPSLongitudeRef']
GPSDATE_CODE         = GPS_TAG_TO_GPSID['GPSDateStamp']
GPSTIME_CODE         = GPS_TAG_TO_GPSID['GPSTimeStamp']


def read_exif_tags(pil_img, exif_tagid_list, default_list=None):
    if default_list is None:
        default_list = [None for _ in range(len(exif_tagid_list))]
    exif_dict = get_exif_dict(pil_img)
    exif_val_list = [exif_dict.get(key, default) for key, default in
                     zip(exif_tagid_list, default_list)]
    return exif_val_list


def get_exif_dict(pil_img):
    """ Returns exif dictionary by TAGID """
    try:
        exif_dict = pil_img._getexif()
        if exif_dict is None:
            raise AttributeError
        assert isinstance(exif_dict, dict), 'type(exif_dict)=%r' % type(exif_dict)
    except (ValueError, IndexError, AttributeError, OverflowError):
        exif_dict = {}
    except Exception as ex:
        ut.printex(ex, 'get_exif_dict failed in an unexpected way')
        raise
    return exif_dict


def get_exif_dict2(pil_img):
    """ Returns exif dictionary by TAG (less efficient)"""
    try:
        exif_dict = pil_img._getexif()
        if exif_dict is None:
            raise AttributeError
        assert isinstance(exif_dict, dict), 'type(exif_dict)=%r' % type(exif_dict)
        exif_dict2 = make_exif_dict_human_readable(exif_dict)
    except AttributeError:
        exif_dict2 = {}
    except OverflowError:
        exif_dict2 = {}
    return exif_dict2


def make_exif_dict_human_readable(exif_dict):
    exif_dict2 = {TAGS.get(key, key): val
                  for (key, val) in exif_dict.items()}
    return exif_dict2


def check_exif_keys(pil_img):
    info_ = pil_img._getexif()
    valid_keys = []
    invalid_keys = []
    for key, val in info_.items():
        try:
            exif_keyval = TAGS[key]
            valid_keys.append((key, exif_keyval))
        except KeyError:
            invalid_keys.append(key)
    print('[io] valid_keys = ' + '\n'.join(valid_keys))
    print('-----------')


def read_all_exif_tags(pil_img):
    info_ = pil_img._getexif()
    exif = {} if info_ is None else {
        TAGS.get(key, key): val
        for key, val in info_.items()
    }
    return exif


def get_exif_tagids(tag_list):
    tagid_list = [EXIF_TAG_TO_TAGID[tag] for tag in tag_list]
    return tagid_list


def read_one_exif_tag(pil_img, tag):
    try:
        exif_key = TAGS.keys()[TAGS.values().index(tag)]
    except ValueError:
        return 'Invalid EXIF Tag'
    info_ = pil_img._getexif()
    if info_ is None:
        return None
    else:
        invalid_str = 'Invalid EXIF Key: exif_key=%r, tag=%r' % (exif_key, tag)
        exif_val = info_.get(exif_key, invalid_str)
    return exif_val


def read_exif(fpath, tag=None):
    try:
        pil_img = Image.open(fpath)
        if not hasattr(pil_img, '_getexif'):
            return 'No EXIF Data'
    except IOError as ex:
        print('Caught IOError: %r' % (ex,))
        image_shared.print_image_checks(fpath)
        raise
        return {} if tag is None else None
    if tag is None:
        exif = read_all_exif_tags(pil_img)
    else:
        exif = read_one_exif_tag(pil_img, tag)
    del pil_img
    return exif


def get_exist(data, key):
    if key in data:
        return data[key]
    return None


def convert_degrees(value):
    """
    Helper function to convert the GPS coordinates stored in the EXIF to degress in float format

    References:
        http://en.wikipedia.org/wiki/Geographic_coordinate_conversion
    """
    d0 = value[0][0]
    d1 = value[0][1]
    d = float(d0) / float(d1)

    m0 = value[1][0]
    m1 = value[1][1]
    m = float(m0) / float(m1)

    s0 = value[2][0]
    s1 = value[2][1]
    s = float(s0) / float(s1)

    degrees_float = d + (m / 60.0) + (s / 3600.0)
    return degrees_float


def get_unixtime_gps(exif_dict, default=-1):
    if GPSINFO_CODE in exif_dict:
        gps_info = exif_dict[GPSINFO_CODE]

        if (GPSDATE_CODE in gps_info and
             GPSTIME_CODE in gps_info):
            gps_date = gps_info[GPSDATE_CODE]
            gps_time = gps_info[GPSTIME_CODE]

            try:
                hour = gps_time[0][0]
                minute = gps_time[1][0]
                second = gps_time[2][0] / 1000.0
                exiftime = '%s %02d:%02d:%02d' % (gps_date, hour, minute, second)

                unixtime = util_time.exiftime_to_unixtime(exiftime)  # convert to unixtime
                return unixtime
            except Exception:
                pass

    return default


def get_lat_lon(exif_dict, default=(-1, -1)):
    r"""
    Returns the latitude and longitude, if available, from the provided
    exif_data2 (obtained through exif_data2 above)

    Notes:
        Might need to downgrade to Pillow 2.9.0 to solve a bug with getting GPS
        https://github.com/python-pillow/Pillow/issues/1477

        python -c "from PIL import Image; print(Image.PILLOW_VERSION)"

        pip uninstall Pillow
        pip install Pillow==2.9

    CommandLine:
        python -m vtool_ibeis.exif --test-get_lat_lon

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.exif import *  # NOQA
        >>> import numpy as np
        >>> image_fpath = ut.grab_file_url('http://images.summitpost.org/original/769474.JPG')
        >>> pil_img = Image.open(image_fpath)
        >>> exif_dict = get_exif_dict(pil_img)
        >>> latlon = get_lat_lon(exif_dict)
        >>> result = np.array_str(np.array(latlon), precision=3)
        >>> print(result)
    """
    try:
        if GPSINFO_CODE in exif_dict:
            gps_info = exif_dict[GPSINFO_CODE]

            if (GPSLATITUDE_CODE in gps_info and
                 GPSLATITUDEREF_CODE in gps_info and
                 GPSLONGITUDE_CODE in gps_info and
                 GPSLONGITUDEREF_CODE in gps_info):
                gps_latitude      = gps_info[GPSLATITUDE_CODE]
                gps_latitude_ref  = gps_info[GPSLATITUDEREF_CODE]
                gps_longitude     = gps_info[GPSLONGITUDE_CODE]
                gps_longitude_ref = gps_info[GPSLONGITUDEREF_CODE]
                try:
                    lat = convert_degrees(gps_latitude)
                    if gps_latitude_ref != 'N':
                        lat = 0 - lat

                    lon = convert_degrees(gps_longitude)
                    if gps_longitude_ref != 'E':
                        lon = 0 - lon
                    return lat, lon
                except (ZeroDivisionError, TypeError):
                    # FIXME: -1, -1 is not a good invalid GPS
                    # Find out what the divide by zero really means
                    # currently we think it just is bad gps data
                    pass
    except Exception:
        pass
    return default


def get_orientation(exif_dict, default=0, on_error='warn'):
    r"""
    Returns the image orientation, if available, from the provided
    exif_data2 (obtained through exif_data2 above)

    CommandLine:
        python -m vtool_ibeis.exif --test-get_orientation

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.exif import *  # NOQA
        >>> from os.path import join
        >>> import numpy as np
        >>> url = 'https://cthulhu.dyn.wildme.io/public/models/orientation.zip'
        >>> images_path = ut.grab_zipped_url(url)
        >>> result = []
        >>> for index in range(3):
        >>>     image_filename = 'orientation_%05d.JPG' % (index + 1, )
        >>>     pil_img = Image.open(join(images_path, image_filename))
        >>>     exif_dict = get_exif_dict(pil_img)
        >>>     orient = get_orientation(exif_dict)
        >>>     result.append(orient)
        >>> print(result)
        [1, 6, 8]
    """
    if ORIENTATION_CODE in exif_dict:
        orient = exif_dict[ORIENTATION_CODE]
        if orient in ORIENTATION_DICT:
            if ORIENTATION_DICT[orient] is None:
                raise NotImplementedError('Orientation not defined')
            default = orient
        else:
            msg = 'Unrecognized orientation=%r in Exif' % (orient,)
            if on_error == 'warn':
                warnings.warn(msg, RuntimeWarning)
            else:
                raise AssertionError(msg)
    return default


def get_orientation_str(exif_dict, **kwargs):
    r"""
    Returns the image orientation strings, if available, from the provided
    exif_data2 (obtained through exif_data2 above)

    CommandLine:
        python -m vtool_ibeis.exif --test-get_orientation_str

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.exif import *  # NOQA
        >>> from os.path import join
        >>> import numpy as np
        >>> url = 'https://cthulhu.dyn.wildme.io/public/models/orientation.zip'
        >>> images_path = ut.grab_zipped_url(url)
        >>> result = []
        >>> for index in range(3):
        >>>     image_filename = 'orientation_%05d.JPG' % (index + 1, )
        >>>     pil_img = Image.open(join(images_path, image_filename))
        >>>     exif_dict = get_exif_dict(pil_img)
        >>>     orient_str = get_orientation_str(exif_dict)
        >>>     result.append(orient_str)
        >>> print(result)
        ['Normal', '90 Clockwise', '90 Counter-Clockwise']
    """
    orient = get_orientation(exif_dict, **kwargs)
    orient_str = ORIENTATION_DICT[orient]
    return orient_str


def get_unixtime(exif_dict, default=-1):
    """
    TODO: Exif.Image.TimeZoneOffset

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.exif import *  # NOQA
        >>> image_fpath = ut.grab_file_url('http://images.summitpost.org/original/769474.JPG')
        >>> pil_img = Image.open(image_fpath)
        >>> exif_dict = get_exif_dict(pil_img)
    """
    exiftime  = exif_dict.get(DATETIMEORIGINAL_TAGID, default)
    if isinstance(exiftime, tuple) and len(exiftime) == 1:
        # hack, idk why
        exiftime = exiftime[0]
    if exiftime != -1 and len(exiftime) == 19 and exiftime[-1] != ' ' and exiftime[-3:-1] == ': ':
        # Hack for weird fluke exif times '2009:10:01 11:52: 1'
        exiftime = list(exiftime)
        exiftime[-2] = '0'
        exiftime = ''.join(exiftime)
    unixtime = util_time.exiftime_to_unixtime(exiftime)  # convert to unixtime
    timezone_offset = exif_dict.get('TimeZoneOffset', None)
    if timezone_offset is not None:
        pass
    return unixtime


def parse_exif_unixtime(image_fpath):
    r"""
    Args:
        image_fpath (str):

    Returns:
        float: unixtime

    CommandLine:
        python -m vtool_ibeis.exif --test-parse_exif_unixtime

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.exif import *  # NOQA
        >>> image_fpath = ut.grab_file_url('http://images.summitpost.org/original/769474.JPG')
        >>> unixtime = parse_exif_unixtime(image_fpath)
        >>> result = str(unixtime)
        >>> print(result)
    """
    pil_img = image_shared.open_pil_image(image_fpath)
    exif_dict = get_exif_dict(pil_img)
    unixtime = get_unixtime(exif_dict)
    return unixtime


def parse_exif_unixtime_gps(image_fpath):
    pil_img = image_shared.open_pil_image(image_fpath)
    exif_dict = get_exif_dict(pil_img)
    unixtime = get_unixtime_gps(exif_dict)
    return unixtime


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.exif
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
Sympy helpers
"""
import numpy as np
import six
import utool as ut
import ubelt as ub


def custom_sympy_attrs(mat):
    import sympy
    def matmul(other, hold=True):
        if hold:
            new = sympy.MatMul(mat, other)
        else:
            new = mat.multiply(other)
        custom_sympy_attrs(new)
        return new
    def inv_():
        new = mat.inv()
        custom_sympy_attrs(new)
        return new
    setattr(mat, 'matmul', matmul)
    setattr(mat, 'inv_', inv_)
    return mat


def sympy_mat(arr):
    import sympy
    mat = sympy.Matrix(arr)
    mat = custom_sympy_attrs(mat)
    return mat


def evalprint(str_, globals_=None, locals_=None, simplify=False):
    import sympy
    if globals_ is None:
        globals_ = ut.get_parent_frame().f_globals
    if locals_ is None:
        locals_ = ut.get_parent_frame().f_locals
    if isinstance(str_, six.string_types):
        var = eval(str_, globals_, locals_)
    else:
        var = str_
        str_ = ut.get_varname_from_stack(var, N=1)
    if simplify is True:
        var = sympy.simplify(var)
    print(ub.hzcat(str_ + ' = ', repr(var)))


def check_expr_eq(expr1, expr2, verbose=True):
    """
    Does not work in general. Problem is not decidable.
    Thanks Richard.

    Args:
        expr1 (?):
        expr2 (?):

    CommandLine:
        python -m vtool_ibeis.symbolic --test-check_expr_eq

    SeeALso:
        vt.symbolic_randcheck

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.symbolic import *  # NOQA
        >>> expr1 = sympy.Matrix([ [sx*x + 1.0*tx + w1*y], [sy*y + 1.0*ty + w2*x], [1.0]])
        >>> expr2 = sympy.Matrix([ [sx*x + tx + w1*y], [sy*y + ty + w2*x], [1]])
        >>> result = check_expr_eq(expr1, expr2)
        >>> print(result)
    """
    import sympy
    if isinstance(expr1, six.string_types):
        expr1 = sympy.simplify(expr1)
    if isinstance(expr2, six.string_types):
        expr2 = sympy.simplify(expr2)
    print(ub.hzcat('Checking if ', repr(expr1), ' == ', repr(expr2)))
    random_point_check = expr1.equals(expr2)
    if random_point_check is None:
        failexpr = expr1.equals(expr2, failing_expression=True)
        print('failexpr = %r' % (failexpr,))
        random_point_check = False
    print('... seems %r' % (random_point_check,))
    #return random_point_check
    expr3 = expr1 - expr2
    if not random_point_check and True:
        common_symbols = expr1.free_symbols.intersection(expr2.free_symbols)
        if len(common_symbols):
            y = sympy.symbols('y')  # Hack, should be a new symbol
            symbol = common_symbols.pop()
            soln1 = sympy.solve(sympy.Eq(sympy.simplify(expr1), y), symbol)
            soln2 = sympy.solve(sympy.Eq(sympy.simplify(expr2), y), symbol)
            print('Solving expr1 for common symbol: ' + str(soln1))
            print('Solving expr2 for common symbol: ' + str(soln2))
            if soln1 == soln2:
                print('This seems True')
            else:
                print('This seems False')
        sympy.solve(sympy.Eq(sympy.simplify(expr2), y), 'd')
    print(ub.hzcat('... checking 0 ', repr(expr3)))
    # Does not always work.
    print('(not gaurenteed to work) expr3.is_zero = %r' % (expr3.is_zero,))
    return expr3.is_zero


def symbolic_randcheck(expr1, expr2, domain={}, n=10):
    def get_domain(key, domain={}, rng=np.random):
        min_, max_ = domain.get(key, (-100, 100))
        range_ = max_ - min_
        return (rng.rand() * (range_)) + min_
    num_checks = n
    input_list = []
    results_list = []
    for num in range(num_checks):
        expr1_subs = {key: get_domain(key, domain) for key in expr1.free_symbols}
        expr2_subs = {key: expr1_subs[key] if key in expr1_subs else get_domain(key, domain)
                      for key in expr2.free_symbols}
        expr1_value = expr1.evalf(subs=expr1_subs)
        expr2_value = expr2.evalf(subs=expr2_subs)
        input_list.append((expr1_subs, expr2_subs))
        results_list.append((expr1_value, expr2_value))
    results_list = np.array(results_list)
    #truth_list = np.allclose(results_list.T[0], results_list.T[1])
    truth_list = results_list.T[0] == results_list.T[1]
    return truth_list, results_list, input_list


def sympy_latex_repr(expr1):
    import sympy
    expr1_repr = sympy.latex(expr1)
    expr1_repr = expr1_repr.replace('\\\\', '\\\\\n')
    expr1_repr = expr1_repr.replace(r'\left[\begin{smallmatrix}{}', '\\MAT{\n')
    expr1_repr = expr1_repr.replace(r'\end{smallmatrix}\right]', '\n}')
    expr1_repr = expr1_repr.replace(r'\left[\begin{matrix}', '\\BIGMAT{\n')
    expr1_repr = expr1_repr.replace(r'\end{matrix}\right]', '\n}')
    expr1_repr = expr1_repr.replace(r'\left (', '(')
    expr1_repr = expr1_repr.replace(r'\right )', ')')
    expr1_repr = expr1_repr.replace(r'\left(', '(')
    expr1_repr = expr1_repr.replace(r'\right)', ')')
    # hack of align
    expr1_repr = ut.align(expr1_repr, '&', pos=None)
    return expr1_repr
    #print(expr1_repr)


def sympy_numpy_repr(expr1):
    import re
    expr1_repr = repr(expr1)
    expr1_repr = expr1_repr.replace('Matrix', 'np.array')
    expr1_repr = re.sub('\\bsin\\b', 'np.sin', expr1_repr)
    expr1_repr = re.sub('\\bcos\\b', 'np.cos', expr1_repr)
    expr1_repr = ut.autoformat_pep8(expr1_repr)
    print(expr1_repr)
    #import autopep8
    #autopep8.fix_code(expr1_repr)


"""
Symbolic Scrap Work:
The number of negative reviews needed is usually much larger than the number of
positive reviews.

import sympy
from sympy.abc import theta
import sympy.stats
from sympy.stats import E as mean
items = sympy.symbols('a, b, c, d')

from sympy.stats import FiniteRV, P, E
density = {0: .1, 1: .2, 2: .3, 3: .4}
X = FiniteRV('X', density)

cs = sympy.stats.FiniteRV(str('X'), {0: .5, 1: .5})

cs = [[None] * np.random.randint(10) for _ in range(1)]
print(sum(len(c) - 1 for c in cs))
print(np.mean([len(c) for c in cs]) * len(cs) - len(cs))

ori = theta
x, y, iv11, iv21, iv22, patch_size = sympy.symbols('x y iv11 iv21 iv22 S')



"""


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.symbolic
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
These functions might be PR quality for numpy.
"""
import numpy as np
import utool as ut  # NOQA
import ubelt as ub  # NOQA


def atleast_nd(arr, n, tofront=False):
    r"""
    View inputs as arrays with at least n dimensions.
    TODO: Submit as a PR to numpy

    Args:
        arr (array_like): One array-like object.  Non-array inputs are
                converted to arrays.  Arrays that already have n or more
                dimensions are preserved.
        n (int): number of dimensions to ensure
        tofront (bool): if True new dimensions are added to the front of the
            array.  otherwise they are added to the back.

    CommandLine:
        python -m vtool_ibeis.numpy_utils atleast_nd

    Returns:
        ndarray :
            An array with ``a.ndim >= n``.  Copies are avoided where possible,
            and views with three or more dimensions are returned.  For example,
            a 1-D array of shape ``(N,)`` becomes a view of shape
            ``(1, N, 1)``, and a 2-D array of shape ``(M, N)`` becomes a view
            of shape ``(M, N, 1)``.

    See Also:
        ensure_shape, np.atleast_1d, np.atleast_2d, np.atleast_3d

    Example0:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.numpy_utils import *  # NOQA
        >>> n = 2
        >>> arr = np.array([1, 1, 1])
        >>> arr_ = atleast_nd(arr, n)
        >>> result = ub.repr2(arr_.tolist())
        >>> print(result)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.numpy_utils import *  # NOQA
        >>> n = 4
        >>> arr1 = [1, 1, 1]
        >>> arr2 = np.array(0)
        >>> arr3 = np.array([[[[[1]]]]])
        >>> arr1_ = atleast_nd(arr1, n)
        >>> arr2_ = atleast_nd(arr2, n)
        >>> arr3_ = atleast_nd(arr3, n)
        >>> result1 = ub.repr2(arr1_.tolist())
        >>> result2 = ub.repr2(arr2_.tolist())
        >>> result3 = ub.repr2(arr3_.tolist())
        >>> result = '\n'.join([result1, result2, result3])
        >>> print(result)
    """
    arr_ = np.asanyarray(arr)
    ndims = len(arr_.shape)
    if n is not None and ndims <  n:
        # append the required number of dimensions to the front or back
        if tofront:
            expander = (None,) * (n - ndims) + (Ellipsis,)
        else:
            expander = (Ellipsis,) + (None,) * (n - ndims)
        arr_ = arr_[expander]
    return arr_


def ensure_shape(arr, dimshape):
    """
    TODO: Submit as a PR to numpy?

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.numpy_utils import *  # NOQA
        >>> ensure_shape(np.array([[1, 2]]), (None, 2))
        >>> ensure_shape(np.array([]), (None, 2))
    """
    if isinstance(dimshape, tuple):
        n = len(dimshape)
    else:
        n = dimshape
        dimshape = None
    arr_ = atleast_nd(arr, n)
    if dimshape is not None:
        newshape = tuple([
            d1 if d2 is None else d2
            for d1, d2 in zip(arr_.shape, dimshape)])
        arr_.shape = newshape
    return arr_


def fromiter_nd(iter_, shape, dtype):
    """
    Like np.fromiter but handles iterators that generated
    n-dimensional arrays. Slightly faster than np.array.

    Note:
        np.vstack(list_) is still faster than
        vt.fromiter_nd(ut.iflatten(list_))

    Args:
        iter_ (iter): an iterable that generates homogenous ndarrays
        shape (tuple): the expected output shape
        dtype (dtype): the numpy datatype of the generated ndarrays

    Note:
        The iterable must yeild a numpy array. It cannot yeild a Python list.

    CommandLine:
        python -m vtool_ibeis.numpy_utils fromiter_nd

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.numpy_utils import *  # NOQA
        >>> dtype = float
        >>> total = 11
        >>> rng = np.random.RandomState(0)
        >>> iter_ = (rng.rand(5, 7, 3) for _ in range(total))
        >>> shape = (total, 5, 7, 3)
        >>> result = fromiter_nd(iter_, shape, dtype)
        >>> assert result.shape == shape

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.numpy_utils import *  # NOQA
        >>> dtype = int
        >>> qfxs = np.array([1, 2, 3])
        >>> dfxs = np.array([4, 5, 6])
        >>> iter_ = (np.array(x) for x in ut.product(qfxs, dfxs))
        >>> total = len(qfxs) * len(dfxs)
        >>> shape = (total, 2)
        >>> result = fromiter_nd(iter_, shape, dtype)
        >>> assert result.shape == shape

    Timeit:
        >>> dtype = np.uint8
        >>> feat_dim = 128
        >>> mu = 1000
        >>> sigma = 500
        >>> n_data = 1000
        >>> rng = np.random.RandomState(42)
        >>> n_feat_list = np.clip(rng.randn(n_data) * sigma + mu, 0, np.inf).astype(int)
        >>> # Make a large list of vectors of various sizes
        >>> print('Making random vectors')
        >>> vecs_list = [(rng.rand(num, feat_dim) * 255).astype(dtype) for num in n_feat_list]
        >>> mega_bytes = sum([x.nbytes for x in vecs_list]) / 2 ** 20
        >>> print('mega_bytes = %r' % (mega_bytes,))
        >>> import itertools as it
        >>> import vtool_ibeis as vt
        >>> n_total = n_feat_list.sum()
        >>> target1 = np.vstack(vecs_list)
        >>> iter_ = it.chain.from_iterable(vecs_list)
        >>> shape = (n_total, feat_dim)
        >>> target2 = vt.fromiter_nd(it.chain.from_iterable(vecs_list), shape, dtype=dtype)
        >>> assert np.all(target1 == target2)

        %timeit np.vstack(vecs_list)
        20.4ms
        %timeit vt.fromiter_nd(it.chain.from_iterable(vecs_list), shape, dtype)
        102ms

        iter_ = it.chain.from_iterable(vecs_list)
        %time vt.fromiter_nd(iter_, shape, dtype)
        %time np.vstack(vecs_list)
    """
    num_rows = shape[0]
    chunksize = np.prod(shape[1:])
    itemsize = np.dtype(dtype).itemsize
    # Create dtype that makes an entire ndarray appear as a single item
    chunk_dtype = np.dtype((np.void, itemsize * chunksize))
    arr = np.fromiter(iter_, count=num_rows, dtype=chunk_dtype)
    # Convert back to original dtype and shape
    arr = arr.view(dtype)
    arr.shape = shape
    return arr


def index_to_boolmask(index_list, maxval=None, isflat=True):
    r"""
    transforms a list of indicies into a boolean mask

    Args:
        index_list (ndarray):
        maxval (None): (default = None)

    Kwargs:
        maxval

    Returns:
        ndarray: mask

    CommandLine:
        python -m vtool_ibeis.util_numpy index_to_boolmask

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_numpy import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> index_list = np.array([(0, 0), (1, 1), (2, 1)])
        >>> maxval = (3, 3)
        >>> mask = vt.index_to_boolmask(index_list, maxval, isflat=False)
        >>> result = ('mask =\n%s' % (str(mask.astype(np.uint8)),))
        >>> print(result)
        [[1 0 0]
         [0 1 0]
         [0 1 0]]

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.util_numpy import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> index_list = np.array([0, 1, 4])
        >>> maxval = 5
        >>> mask = vt.index_to_boolmask(index_list, maxval, isflat=True)
        >>> result = ('mask = %s' % (str(mask.astype(np.uint8)),))
        >>> print(result)
        mask = [1 1 0 0 1]

    """
    #assert index_list.min() >= 0
    if maxval is None:
        maxval = index_list.max()
    mask = np.zeros(maxval, dtype=bool)
    if not isflat:
        # assumes non-flat
        mask.__setitem__(tuple(index_list.T), True)
        #mask.__getitem__(tuple(index_list.T))
    else:
        mask[index_list] = True
    return mask


def multiaxis_reduce(ufunc, arr, startaxis=0):
    """
    used to get max/min over all axes after <startaxis>

    CommandLine:
        python -m vtool_ibeis.numpy_utils --test-multiaxis_reduce

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.numpy_utils import *  # NOQA
        >>> rng = np.random.RandomState(0)
        >>> arr = (rng.rand(4, 3, 2, 1) * 255).astype(np.uint8)
        >>> ufunc = np.amax
        >>> startaxis = 1
        >>> out_ = multiaxis_reduce(ufunc, arr, startaxis)
        >>> result = out_
        >>> print(result)
        [182 245 236 249]
    """
    num_iters = len(arr.shape) - startaxis
    out_ = ufunc(arr, axis=startaxis)
    for _ in range(num_iters - 1):
        out_ = ufunc(out_, axis=1)
    return out_


def iter_reduce_ufunc(ufunc, arr_iter, out=None):
    """
    constant memory iteration and reduction

    applys ufunc from left to right over the input arrays

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.numpy_utils import *  # NOQA
        >>> arr_list = [
        ...     np.array([0, 1, 2, 3, 8, 9]),
        ...     np.array([4, 1, 2, 3, 4, 5]),
        ...     np.array([0, 5, 2, 3, 4, 5]),
        ...     np.array([1, 1, 6, 3, 4, 5]),
        ...     np.array([0, 1, 2, 7, 4, 5])
        ... ]
        >>> memory = np.array([9, 9, 9, 9, 9, 9])
        >>> gen_memory = memory.copy()
        >>> def arr_gen(arr_list, gen_memory):
        ...     for arr in arr_list:
        ...         gen_memory[:] = arr
        ...         yield gen_memory
        >>> print('memory = %r' % (memory,))
        >>> print('gen_memory = %r' % (gen_memory,))
        >>> ufunc = np.maximum
        >>> res1 = iter_reduce_ufunc(ufunc, iter(arr_list), out=None)
        >>> res2 = iter_reduce_ufunc(ufunc, iter(arr_list), out=memory)
        >>> res3 = iter_reduce_ufunc(ufunc, arr_gen(arr_list, gen_memory), out=memory)
        >>> print('res1       = %r' % (res1,))
        >>> print('res2       = %r' % (res2,))
        >>> print('res3       = %r' % (res3,))
        >>> print('memory     = %r' % (memory,))
        >>> print('gen_memory = %r' % (gen_memory,))
        >>> assert np.all(res1 == res2)
        >>> assert np.all(res2 == res3)
    """
    # Get first item in iterator
    try:
        initial = next(arr_iter)
    except StopIteration:
        return None
    # Populate the outvariable if specified otherwise make a copy of the first
    # item to be the output memory
    if out is not None:
        out[:] = initial
    else:
        out = initial.copy()
    # Iterate and reduce
    for arr in arr_iter:
        ufunc(out, arr, out=out)
    return out


def unique_row_indexes(arr):
    """ np.unique on rows

    Args:
        arr (ndarray): 2d array

    Returns:
        ndarray: unique_rowx

    References:
        http://stackoverflow.com/questions/16970982/find-unique-rows-in-numpy-array

    CommandLine:
        python -m vtool_ibeis.numpy_utils --test-unique_row_indexes

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.numpy_utils import *  # NOQA
        >>> arr = np.array([[0, 0], [0, 1], [1, 0], [1, 1], [0, 0], [.534, .432], [.534, .432], [1, 0], [0, 1]])
        >>> unique_rowx = unique_row_indexes(arr)
        >>> result = ('unique_rowx = %s' % (ub.repr2(unique_rowx),))
        >>> print(result)
        unique_rowx = np.array([0, 1, 2, 3, 5], dtype=np.int64)

    Ignore:
        %timeit unique_row_indexes(arr)
        %timeit compute_unique_data_ids(arr)
        %timeit compute_unique_integer_data_ids(arr)

    """
    void_dtype = np.dtype((np.void, arr.dtype.itemsize * arr.shape[1]))
    arr_void_view = np.ascontiguousarray(arr).view(void_dtype)
    _, unique_rowx = np.unique(arr_void_view, return_index=True)
    # cast back to original dtype
    unique_rowx.sort()
    return unique_rowx


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.numpy_utils
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# -*- coding: utf-8 -*-
r"""
TODO: Look at this file
    http://www.lfd.uci.edu/~gohlke/code/transformations.py.html

# Sympy:
#     >>> # https://groups.google.com/forum/#!topic/sympy/k1HnZK_bNNA
#     >>> import vtool_ibeis as vt
#     >>> import sympy
#     >>> from sympy.abc import theta
#     >>> x, y, a, c, d, sx, sy  = sympy.symbols('x y a c d, sx, sy')
#     >>> R = vt.sympy_mat(vt.rotation_mat3x3(theta, sin=sympy.sin, cos=sympy.cos))
#     >>> vt.evalprint('R')
#     >>> #evalprint('R.inv()')
#     >>> vt.evalprint('sympy.simplify(R.inv())')
#     >>> #evalprint('sympy.simplify(R.inv().subs(theta, 4))')
#     >>> #print('-------')
#     >>> #invR = sympy_mat(vt.rotation_mat3x3(-theta, sin=sympy.sin, cos=sympy.cos))
#     >>> #evalprint('invR')
#     >>> #evalprint('invR.inv()')
#     >>> #evalprint('sympy.simplify(invR)')
#     >>> #evalprint('sympy.simplify(invR.subs(theta, 4))')
#     >>> print('-------')
#     >>> T = vt.sympy_mat(vt.translation_mat3x3(x, y, None))
#     >>> vt.evalprint('T')
#     >>> vt.evalprint('T.inv()')
#     >>> print('-------')
#     >>> S = vt.sympy_mat(vt.scale_mat3x3(sx, sy, dtype=None))
#     >>> vt.evalprint('S')
#     >>> vt.evalprint('S.inv()')
#     >>> print('-------')
#     >>> print('LaTeX')
#     >>> print(xdev.align('\\\\\n'.join(sympy.latex(R).split(r'\\')).replace('{matrix}', '{matrix}\n'), '&')

"""
import cv2
import numpy as np
import numpy.linalg as npl
import ubelt as ub
import warnings  # NOQA
from .util_math import TAU
from operator import matmul as matrix_multiply  # NOQA


TRANSFORM_DTYPE = np.float64

# Function which multiplies many matrices at once
# from numpy.core.umath_tests import matrix_multiply  # NOQA


def svd(M):
    r"""
    Args:
        M (ndarray): must be either float32 or float64

    Returns:
        tuple: (U, s, Vt)

    CommandLine:
        python -m vtool_ibeis.linalg --test-svd

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> # build test data
        >>> M = np.array([1, 2, 3], dtype=np.float32)
        >>> M = np.array([[20.5812, 0], [3.615, 17.1295]], dtype=np.float64)
        >>> # execute function
        >>> (U, s, Vt) = svd(M)

    Timeit::
        flags = cv2.SVD_FULL_UV
        %timeit cv2.SVDecomp(M, flags=flags)
        %timeit npl.svd(M)
    """
    # V is actually Vt
    flags = cv2.SVD_FULL_UV
    S, U, Vt = cv2.SVDecomp(M, flags=flags)
    s = S.flatten()
    #U, s, Vt = npl.svd(M)
    return U, s, Vt


def gauss2d_pdf(x_, y_, sigma=None, mu=None):
    """
    Input: x and y coordinate of a 2D gaussian
           sigma, mu - covariance and mean vector
    Output: The probability density at that point
    """
    if sigma is None:
        sigma = np.eye(2)
    else:
        if not isinstance(sigma, np.ndarray):
            sigma = np.eye(2) * sigma
    if mu is None:
        mu = np.array([0, 0])
    x = np.array([x_, y_])
    size = len(x)
    if size == len(mu) and (size, size) == sigma.shape:
        det = npl.det(sigma)
        if det == 0:
            raise NameError('The covariance matrix cant be singular')
    denom1 = TAU ** (size / 2.0)
    denom2 = np.sqrt(det)
    #norm_const = 1.0 / (denom1 * denom2)
    norm_const = np.reciprocal(denom1 * denom2)
    x_mu = x - mu  # deviation from mean
    invSigma = npl.inv(sigma)  # inverse covariance
    exponent = -0.5 * (x_mu.dot(invSigma).dot(x_mu.T))
    result = norm_const * np.exp(exponent)
    return result


def rotation_mat3x3(radians, sin=np.sin, cos=np.cos):
    """
    References:
        https://en.wikipedia.org/wiki/Rotation_matrix
    """
    # TODO: handle array inputs
    sin_ = sin(radians)
    cos_ = cos(radians)
    R = np.array(((cos_, -sin_,  0),
                  (sin_,  cos_,  0),
                  (   0,     0,  1),))
    return R


def rotation_mat2x2(theta):
    sin_ = np.sin(theta)
    cos_ = np.cos(theta)
    rot_ = np.array(((cos_, -sin_),
                     (sin_,  cos_),))
    return rot_


def transform_around(M, x, y):
    """ translates to origin, applies transform and then translates back """
    tr1_ = translation_mat3x3(-x, -y)
    tr2_ = translation_mat3x3(x, y)
    M_ = tr2_.dot(M).dot(tr1_)
    return M_


def rotation_around_mat3x3(theta, x0, y0, x1=None, y1=None):
    if x1 is None:
        x1 = x0
    if y1 is None:
        y1 = y0
    # rot = rotation_mat3x3(theta)
    # return transform_around(rot, x, y)
    tr1_ = translation_mat3x3(-x0, -y0)
    rot_ = rotation_mat3x3(theta)
    tr2_ = translation_mat3x3(x1, y1)
    rot = tr2_.dot(rot_).dot(tr1_)
    return rot


def scale_around_mat3x3(sx, sy, x, y):
    scale_ = scale_mat3x3(sx, sy)
    return transform_around(scale_, x, y)


def rotation_around_bbox_mat3x3(theta, bbox0, bbox1=None):
    x, y, w, h = bbox0
    centerx0 = x + (w / 2)
    centery0 = y + (h / 2)
    if bbox1 is None:
        centerx1 = None
        centery1 = None
    else:
        x, y, w, h = bbox1
        centerx1 = x + (w / 2)
        centery1 = y + (h / 2)
    return rotation_around_mat3x3(theta, centerx0, centery0, x1=centerx1, y1=centery1)


def translation_mat3x3(x, y, dtype=TRANSFORM_DTYPE):
    T = np.array([[1, 0,  x],
                  [0, 1,  y],
                  [0, 0,  1]], dtype=dtype)
    return T


def scale_mat3x3(sx, sy=None, dtype=TRANSFORM_DTYPE):
    sy = sx if sy is None else sy
    S = np.array([[sx, 0, 0],
                  [0, sy, 0],
                  [0,  0, 1]], dtype=dtype)
    return S


def shear_mat3x3(shear_x, shear_y, dtype=TRANSFORM_DTYPE):
    shear = np.array([[      1, shear_x, 0],
                      [shear_y,       1, 0],
                      [      0,       0, 1]], dtype=dtype)
    return shear


def affine_mat3x3(sx=1, sy=1, theta=0, shear=0, tx=0, ty=0, trig=np):
    r"""
    Args:
        sx (float): x scale factor (default = 1)
        sy (float): y scale factor (default = 1)
        theta (float): rotation angle (radians) in counterclockwise direction
        shear (float): shear angle (radians) in counterclockwise directions
        tx (float): x-translation (default = 0)
        ty (float): y-translation (default = 0)

    References:
        https://github.com/scikit-image/scikit-image/blob/master/skimage/transform/_geometric.py
    """
    sin1_ = trig.sin(theta)
    cos1_ = trig.cos(theta)
    sin2_ = trig.sin(theta + shear)
    cos2_ = trig.cos(theta + shear)
    Aff = np.array([
        [sx * cos1_, -sy * sin2_, tx],
        [sx * sin1_,  sy * cos2_, ty],
        [        0,            0,  1]
    ])
    return Aff


def affine_around_mat3x3(x, y, sx=1.0, sy=1.0, theta=0.0, shear=0.0, tx=0.0,
                         ty=0.0, x2=None, y2=None):
    r"""
    Executes an affine transform around center point (x, y).
    Equivalent to translation.dot(affine).dot(inv(translation))

    Args:
        x (float): center x location in input space
        y (float):  center y location in input space
        sx (float): x scale factor (default = 1)
        sy (float): y scale factor (default = 1)
        theta (float): counter-clockwise rotation angle in radians(default = 0)
        shear (float): counter-clockwise shear angle in radians(default = 0)
        tx (float): x-translation (default = 0)
        ty (float): y-translation (default = 0)
        x2 (float, optional): center y location in output space (default = x)
        y2 (float, optional): center y location in output space (default = y)

    CommandLine:
        python -m vtool_ibeis.linalg affine_around_mat3x3 --show
    CommandLine:
        xdoctest -m ~/code/vtool_ibeis/vtool_ibeis/linalg.py affine_around_mat3x3

    Example:
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> orig_pts = np.array(vt.verts_from_bbox([10, 10, 20, 20]))
        >>> x, y = vt.bbox_center(vt.bbox_from_verts(orig_pts))
        >>> sx, sy = 0.5, 1.0
        >>> theta = 1 * np.pi / 4
        >>> shear = .1 * np.pi / 4
        >>> tx, ty = 5, 0
        >>> x2, y2 = None, None
        >>> Aff = affine_around_mat3x3(x, y, sx, sy, theta, shear,
        >>>                            tx, ty, x2, y2)
        >>> trans_pts = vt.transform_points_with_homography(Aff, orig_pts.T).T
        >>> # xdoc: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.ensureqt()
        >>> pt.plt.plot(x, y, 'bx', label='center')
        >>> pt.plt.plot(orig_pts.T[0], orig_pts.T[1], 'b-', label='original')
        >>> pt.plt.plot(trans_pts.T[0], trans_pts.T[1], 'r-', label='transformed')
        >>> pt.plt.legend()
        >>> pt.plt.title('Demo of affine_around_mat3x3')
        >>> pt.plt.axis('equal')
        >>> pt.plt.xlim(0, 40)
        >>> pt.plt.ylim(0, 40)
        >>> pt.show_if_requested()

    # Timeit:
    #     >>> from vtool_ibeis.linalg import *  # NOQA
    #     >>> x, y, sx, sy, theta, shear, tx, ty, x2, y2 = (
    #     >>>     256.0, 256.0, 1.5, 1.0, 0.78, 0.2, 0, 100, 500.0, 500.0)
    #     >>> for timer in ub.Timerit(1000, 'old'):  # 19.0697 µs
    #     >>>     with timer:
    #     >>>         tr1_ = translation_mat3x3(-x, -y)
    #     >>>         Aff_ = affine_mat3x3(sx, sy, theta, shear, tx, ty)
    #     >>>         tr2_ = translation_mat3x3(x2, y2)
    #     >>>         Aff1 = tr2_.dot(Aff_).dot(tr1_)
    #     >>> for timer in ub.Timerit(1000, 'new'):  # 11.0242 µs
    #     >>>     with timer:
    #     >>>         Aff2 = affine_around_mat3x3(x, y, sx, sy, theta, shear,
    #     >>>                                     tx, ty, x2, y2)
    #     >>> assert np.all(np.isclose(Aff2, Aff1))

    # Sympy:
    #     >>> from vtool_ibeis.linalg import *  # NOQA
    #     >>> import vtool_ibeis as vt
    #     >>> import sympy
    #     >>> # Shows the symbolic construction of the code
    #     >>> # https://groups.google.com/forum/#!topic/sympy/k1HnZK_bNNA
    #     >>> from sympy.abc import theta
    #     >>> x, y, sx, sy, theta, shear, tx, ty, x2, y2 = sympy.symbols(
    #     >>>     'x, y, sx, sy, theta, shear, tx, ty, x2, y2')
    #     >>> theta = sx = sy = tx = ty = 0
    #     >>> # move to center xy, apply affine transform, move center xy2
    #     >>> tr1_ = translation_mat3x3(-x, -y, dtype=None)
    #     >>> Aff_ = affine_mat3x3(sx, sy, theta, shear, tx, ty, trig=sympy)
    #     >>> tr2_ = translation_mat3x3(x2, y2, dtype=None)
    #     >>> # combine transformations
    #     >>> Aff = vt.sympy_mat(tr2_.dot(Aff_).dot(tr1_))
    #     >>> vt.evalprint('Aff')
    #     >>> print('-------')
    #     >>> print('Numpy')
    #     >>> vt.sympy_numpy_repr(Aff)
    """
    x2 = x if x2 is None else x2
    y2 = y if y2 is None else y2
    # Make auxially varables to reduce the number of sin/cosine calls
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)
    cos_shear_p_theta = np.cos(shear + theta)
    sin_shear_p_theta = np.sin(shear + theta)
    tx_ = -sx * x * cos_theta + sy * y * sin_shear_p_theta + tx + x2
    ty_ = -sx * x * sin_theta - sy * y * cos_shear_p_theta + ty + y2
    # Sympy compiled expression
    Aff = np.array([[sx * cos_theta, -sy * sin_shear_p_theta, tx_],
                    [sx * sin_theta,  sy * cos_shear_p_theta, ty_],
                    [             0,                       0, 1]])
    return Aff


# Ensure that a feature doesn't have multiple assignments
# --------------------------------
# Linear algebra functions on lower triangular matrices


def det_ltri(ltri):
    """ Lower triangular determinant """
    det = ltri[0] * ltri[2]
    return det


def inv_ltri(ltri, det):
    """ Lower triangular inverse """
    inv_ltri = np.array((ltri[2], -ltri[1], ltri[0]), dtype=ltri.dtype) / det
    return inv_ltri


def dot_ltri(ltri1, ltri2):
    """ Lower triangular dot product """
    # use m, n, and o as temporary matrixes
    m11, m21, m22 = ltri1
    n11, n21, n22 = ltri2
    o11 = (m11 * n11)
    o21 = (m21 * n11) + (m22 * n21)
    o22 = (m22 * n22)
    ltri3 = np.array((o11, o21, o22), dtype=ltri1.dtype)
    return ltri3


def whiten_xy_points(xy_m):
    """
    whitens points to mean=0, stddev=1 and returns transformation

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> from vtool_ibeis import demodata
        >>> xy_m = demodata.get_dummy_xy()
        >>> tup = whiten_xy_points(xy_m)
        >>> xy_norm, T = tup
        >>> result = (ub.hash_data(tup))
        >>> print(result)
    """
    mu_xy  = xy_m.mean(1)  # center of mass
    std_xy = xy_m.std(1)
    std_xy[std_xy == 0] = 1  # prevent divide by zero
    tx, ty = -mu_xy / std_xy
    sx, sy = 1 / std_xy
    T = np.array([(sx, 0, tx),
                  (0, sy, ty),
                  (0,  0,  1)])
    xy_norm = ((xy_m.T - mu_xy) / std_xy).T
    return xy_norm, T


def add_homogenous_coordinate(_xys):
    r"""
    CommandLine:
        python -m vtool_ibeis.linalg --test-add_homogenous_coordinate

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> _xys = np.array([[ 2.,  0.,  0.,  2.],
        ...                  [ 2.,  2.,  0.,  0.]], dtype=np.float32)
        >>> _xyzs = add_homogenous_coordinate(_xys)
        >>> assert np.all(_xys == remove_homogenous_coordinate(_xyzs))
        >>> result = ub.repr2(_xyzs, with_dtype=True)
        >>> print(result)
    """
    assert _xys.shape[0] == 2
    _zs = np.ones((1, _xys.shape[1]), dtype=_xys.dtype)
    _xyzs = np.vstack((_xys, _zs))
    return _xyzs


def remove_homogenous_coordinate(_xyzs):
    r"""
    normalizes 3d homogonous coordinates into 2d coordinates

    Args:
        _xyzs (ndarray): of shape (3, N)

    Returns:
        ndarray: _xys of shape (2, N)

    CommandLine:
        python -m vtool_ibeis.linalg --test-remove_homogenous_coordinate

    Example0:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> _xyzs = np.array([[ 2.,   0.,  0.,  2.],
        ...                   [ 2.,   2.,  0.,  0.],
        ...                   [ 1.2,  1.,  1.,  2.]], dtype=np.float32)
        >>> _xys = remove_homogenous_coordinate(_xyzs)
        >>> result = ub.repr2(_xys, precision=3, with_dtype=True)
        >>> print(result)

    Example1:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> _xyzs = np.array([[ 140.,  167.,  185.,  185.,  194.],
        ...                   [ 121.,  139.,  156.,  155.,  163.],
        ...                   [  47.,   56.,   62.,   62.,   65.]])
        >>> _xys = remove_homogenous_coordinate(_xyzs)
        >>> result = ub.repr2(_xys, precision=3)
        >>> print(result)
    """
    assert _xyzs.shape[0] == 3
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        _xys = np.divide(_xyzs[0:2], _xyzs[None, 2])
    return _xys


def transform_points_with_homography(H, _xys):
    """
    Args:
        H (ndarray[float64_t, ndim=2]):  homography/perspective matrix
        _xys (ndarray[ndim=2]): (2 x N) array
    """
    xyz  = add_homogenous_coordinate(_xys)
    xyz_t = H @ xyz
    xy_t  = remove_homogenous_coordinate(xyz_t)
    return xy_t


def normalize_rows(arr, out=None):
    """ DEPRICATE """
    assert len(arr.shape) == 2
    return normalize(arr, axis=1, out=out)


def normalize(arr, ord=None, axis=None, out=None):
    r"""
    Returns all row vectors normalized by their magnitude.

    Args:
        arr (ndarray): row vectors to normalize
        ord (int): type of norm to use (defaults to 2-norm)
            {non-zero int, inf, -inf}
        axis (int): axis to normalize
        out (ndarray): preallocated output

    SeeAlso:
        np.linalg.norm

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> arr = np.array([[1, 2, 3, 4, 5], [2, 2, 2, 2, 2]])
        >>> arr_normed = normalize(arr, axis=1)
        >>> result = ub.hzcat(['arr_normed = ', ub.repr2(arr_normed, precision=2, with_dtype=True)])
        >>> assert np.allclose((arr_normed ** 2).sum(axis=1), [1, 1])
        >>> print(result)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> arr = np.array([ 0.6,  0.1, -0.6])
        >>> arr_normed = normalize(arr)
        >>> result = ub.hzcat(['arr_normed = ', ub.repr2(arr_normed, precision=2)])
        >>> assert np.allclose((arr_normed ** 2).sum(), [1])
        >>> print(result)

    Example:
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> ord_list = [0, 1, 2, np.inf, -np.inf]
        >>> arr = np.array([ 0.6,  0.1, -0.5])
        >>> normed = [(ord, normalize(arr, ord=ord)) for ord in ord_list]
        >>> result = ub.repr2(normed, precision=2, with_dtype=True)
        >>> print(result)
    """
    norm_ = np.linalg.norm(arr, ord=ord, axis=axis, keepdims=True)
    arr_normed = np.divide(arr, norm_, out=out)
    return arr_normed


def random_affine_args(zoom_pdf=None,
                       tx_pdf=None,
                       ty_pdf=None,
                       shear_pdf=None,
                       theta_pdf=None,
                       enable_flip=False,
                       enable_stretch=False,
                       default_distribution='uniform',
                       scalar_anchor='reflect',  # 0
                       txy_pdf=None,
                       rng=np.random):
    r"""
    TODO: allow for a pdf of ranges for each dimension

    If pdfs are tuples it is interpreted as a default (uniform) distribution between the
    two points. A single scalar is a default distribution between -scalar and
    scalar.

    Args:
        zoom_range (tuple): (default = (1.0, 1.0))
        tx_range (tuple): (default = (0.0, 0.0))
        ty_range (tuple): (default = (0.0, 0.0))
        shear_range (tuple): (default = (0, 0))
        theta_range (tuple): (default = (0, 0))
        enable_flip (bool): (default = False)
        enable_stretch (bool): (default = False)
        rng (module):  random number generator(default = numpy.random)

    Returns:
        tuple: affine_args

    CommandLine:
        xdoctest -m ~/code/vtool_ibeis/vtool_ibeis/linalg.py random_affine_args

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.linalg import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> zoom_range = (0.9090909090909091, 1.1)
        >>> tx_pdf = (0.0, 4.0)
        >>> ty_pdf = (0.0, 4.0)
        >>> shear_pdf = (0, 0)
        >>> theta_pdf = (0, 0)
        >>> enable_flip = False
        >>> enable_stretch = False
        >>> rng = np.random.RandomState(0)
        >>> affine_args = random_affine_args(
        >>>     zoom_range, tx_pdf, ty_pdf, shear_pdf, theta_pdf,
        >>>     enable_flip, enable_stretch, rng=rng)
        >>> print('affine_args = %s' % (ub.repr2(affine_args),))
        >>> (sx, sy, theta, shear, tx, ty) = affine_args
        >>> Aff = vt.affine_mat3x3(sx, sy, theta, shear, tx, ty)
        >>> result = ub.repr2(Aff, precision=3, nl=1, with_dtype=0)
        >>> print(result)
        np.array([[ 1.009, -0.   ,  1.695],
                  [ 0.   ,  1.042,  2.584],
                  [ 0.   ,  0.   ,  1.   ]])
    """
    if zoom_pdf is None:
        sx = sy = 1.0
    else:
        log_zoom_range = [np.log(z) for z in zoom_pdf]

        if enable_stretch:
            sx = sy = np.exp(rng.uniform(*log_zoom_range))
        else:
            sx = np.exp(rng.uniform(*log_zoom_range))
            sy = np.exp(rng.uniform(*log_zoom_range))

    def param_distribution(param_pdf, rng=rng):
        if param_pdf is None:
            param = 0
        elif not ub.iterable(param_pdf):
            max_param = param_pdf
            if scalar_anchor == 'reflect':
                min_param = -max_param
            elif scalar_anchor == 0:
                min_param = 0
            param = rng.uniform(min_param, max_param)
        elif isinstance(param_pdf, tuple):
            min_param, max_param = param_pdf
            param = rng.uniform(min_param, max_param)
        else:
            assert False
        return param

    theta = param_distribution(theta_pdf)
    shear = param_distribution(shear_pdf)
    if txy_pdf is not None:
        assert tx_pdf is None, 'cannot specify both'
        assert ty_pdf is None, 'cannot specify both'
        xy_locs, xy_probs = txy_pdf
        tx = param_distribution(tx_pdf)
    else:
        tx = param_distribution(tx_pdf)
        ty = param_distribution(ty_pdf)

    flip = enable_flip and (rng.randint(2) > 0)  # flip half of the time
    if flip:
        # shear 180 degrees + rotate 180 == flip
        theta += np.pi
        shear += np.pi

    affine_args = (sx, sy, theta, shear, tx, ty)
    return affine_args
    #Aff = vt.affine_mat3x3(sx, sy, theta, shear, tx, ty)
    #return Aff


def random_affine_transform(*args, **kwargs):
    affine_args = random_affine_args(**kwargs)
    Aff = affine_mat3x3(*affine_args)
    return Aff


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.linalg
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# LICENCE
import numpy as np
from .util_math import TAU


def atan2(y, x):
    """
    does atan2 but returns from 0 to TAU

    Example:
        >>> from vtool_ibeis.trig import *  # NOQA
        >>> import utool
        >>> rng = np.random.RandomState(0)
        >>> y = rng.rand(1000).astype(np.float64)
        >>> x = rng.rand(1000).astype(np.float64)
        >>> theta = atan2(y, x)
        >>> assert np.all(theta >= 0)
        >>> assert np.all(theta < 2 * np.pi)
        >>> import ubelt as ub
        >>> print(ub.hash_data(theta))

        90fe55311562f1c3ae4...
    """
    theta = np.arctan2(y, x)  # outputs from -TAU/2 to TAU/2
    theta[theta < 0] = theta[theta < 0] + TAU  # map to 0 to TAU (keep coords)
    return theta



"""
abstract which pyflann implementation is used

from vtool_ibeis._pyflann_backend import pyflann
"""
# import ubelt as ub
# import os

__all__ = ['pyflann', 'FLANN_CLS']

FLANN_CLS = None
pyflann = None

try:
    import pyflann_ibeis as pyflann
    FLANN_CLS = pyflann.FLANN
except ImportError:
    FLANN_CLS = None
    pyflann = None
    # try:
    #     import pyflann
    #     FLANN_CLS = pyflann.FLANN
    # except ImportError:
    #     print('no pyflann, using cv2.flann_Index')
    #     try:
    #         import cv2
    #     except Exception:
    #         # print('no pyflann, using dummy index')
    #         class _DUMMY_FLANN_CLS:
    #             def __init__(self):
    #                 raise RuntimeError('pyflann_ibeis is not installed')
    #         FLANN_CLS = _DUMMY_FLANN_CLS
    #     else:
    #         class _CV2_FLANN_CLS:
    #             def __init__(self):
    #                 self._internal = cv2.flann_Index()
    #                 self.params = {}

    #             def build_index(self, features, **flann_params):
    #                 # self._internal.build(features, flann_params, distType)
    #                 self._internal.build(features, flann_params)

    #             def save_index(self, fpath):
    #                 # self._internal.build(features, flann_params, distType)
    #                 self._internal.save(fpath)

    #             def nn_index(self, query, num_neighbors, checks=ub.NoParam):
    #                 # knnSearch(query, knn[, indices[, dists[, params]]]) -> indices, dists
    #                 return self._internal.knnSearch(query, knn=num_neighbors)

    #         FLANN_CLS = _CV2_FLANN_CLS

# print('VTOOL_IBEIS BACKEND FOR pyflann = {!r}'.format(pyflann))
# print('VTOOL_IBEIS BACKEND FOR FLANN_CLS = {!r}'.format(FLANN_CLS))



import numpy as np
import cv2
import utool as ut
import ubelt as ub

DEBUG_SEGM = False


def printDBG(msg):
    if DEBUG_SEGM:
        print(msg)
    pass


def resize_img_and_bbox(img_fpath, bbox_, new_size=None, sqrt_area=400.0):
    printDBG('[segm] imread(%r) ' % img_fpath)
    full_img = cv2.imread(img_fpath)
    (full_h, full_w) = full_img.shape[:2]                 # Image Shape
    printDBG('[segm] full_img.shape=%r' % (full_img.shape,))
    (rw_, rh_) = bbox_[2:]
    # Ensure that we know the new chip size
    if new_size is None:
        target_area = float(sqrt_area) ** 2

        def _resz(w, h):
            ht = np.sqrt(target_area * h / w)
            wt = w * ht / h
            return (int(round(wt)), int(round(ht)))
        new_size_ = _resz(rw_, rh_)
    else:
        new_size_ = new_size
    # Get Scale Factors
    fx = new_size_[0] / rw_
    fy = new_size_[1] / rh_
    printDBG('[segm] fx=%r fy=%r' % (fx, fy))
    dsize = (int(round(fx * full_w)), int(round(fy * full_h)))
    printDBG('[segm] dsize=%r' % (dsize,))
    # Resize the image
    img_resz = cv2.resize(full_img, dsize, interpolation=cv2.INTER_LANCZOS4)
    # Get new ANNOTATION in resized image
    bbox_resz = np.array(np.round(bbox_ * fx), dtype=np.int64)
    return img_resz, bbox_resz


def clean_mask(mask, num_dilate=3, num_erode=3, window_frac=.025):
    """
    Clean the mask
    (num_erode, num_dilate) = (1, 1)
    (w, h) = (10, 10)
    """
    w = h = int(round(min(mask.shape) * window_frac))
    element = cv2.getStructuringElement(cv2.MORPH_CROSS, (w, h))
    _mask = mask
    # compute the closing
    for ix in range(num_dilate):
        _mask = cv2.dilate(_mask, element)
    for ix in range(num_erode):
        _mask = cv2.erode(_mask, element)
    return _mask


def fill_holes(mask):
    mode = cv2.RETR_CCOMP
    method = cv2.CHAIN_APPROX_SIMPLE
    image, contours, hierarchy = cv2.findContours(mask, mode, method)
    out = cv2.drawContours(image, contours, -1, (1, 0, 0))
    return out


def demo_grabcut(bgr_img):
    r"""
    Args:
        img (ndarray[uint8_t, ndim=2]):  image data

    CommandLine:
        python -m vtool_ibeis.segmentation --test-demo_grabcut --show

    SeeAlso:
        python -m ibeis.algo.preproc.preproc_probchip --test-postprocess_dev

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.segmentation import *  # NOQA
        >>> # build test data
        >>> import utool as ut
        >>> import plottool_ibeis as pt
        >>> import vtool_ibeis as vt
        >>> img_fpath = ut.grab_test_imgpath('easy1.png')
        >>> bgr_img = vt.imread(img_fpath)
        >>> # execute function
        >>> print(bgr_img.shape)
        >>> result = demo_grabcut(bgr_img)
        >>> # verify results
        >>> print(result)
        >>> ## xdoctest: +REQUIRES(--show)
        >>> pt.show_if_requested()
    """
    import plottool_ibeis as pt
    from plottool_ibeis import interact_impaint
    label_colors = [       255,           170,            50,          0]
    label_values = [cv2.GC_FGD, cv2.GC_PR_FGD, cv2.GC_PR_BGD, cv2.GC_BGD]
    h, w = bgr_img.shape[0:2]
    init_mask = np.zeros((h, w), dtype=np.float32)  # Initialize: mask
    # Set inside to cv2.GC_PR_FGD (probably forground)
    init_mask[ :, :] = label_colors[label_values.index(cv2.GC_PR_BGD)]
    # Set border to cv2.GC_BGD (definitely background)
    init_mask[ 0, :] = label_colors[label_values.index(cv2.GC_BGD)]
    init_mask[-1, :] = label_colors[label_values.index(cv2.GC_BGD)]
    init_mask[:,  0] = label_colors[label_values.index(cv2.GC_BGD)]
    init_mask[:, -1] = label_colors[label_values.index(cv2.GC_BGD)]
    #import vtool_ibeis as vt
    cached_mask_fpath = 'tmp_mask.png'
    if ub.argflag('--nocache'):
        ut.delete(cached_mask_fpath)
    print('unique init mask colors')
    print(np.unique(init_mask))
    custom_mask = interact_impaint.cached_impaint(bgr_img, cached_mask_fpath,
                                                  label_colors=label_colors,
                                                  init_mask=init_mask)
    print('unique custom mask colors')
    print(np.unique(custom_mask))
    print('delete tmp_mask.png to redo')
    #if ut.checkpath(cached_mask_fpath):
    #    custom_mask = vt.imread(cached_mask_fpath, grayscale=True)
    #else:
    #    custom_mask = interact_impaint.impaint_mask(bgr_img, label_colors, init_mask=init_mask)
    #    vt.imwrite(cached_mask_fpath, custom_mask)

    prior_mask = custom_mask.copy()

    # Convert colors to out labels
    label_locs = [custom_mask == color for color in label_colors]
    # Put user labels in there
    for label_loc, value in zip(label_locs, label_values):
        prior_mask[label_loc] = value
    prior_mask = prior_mask.astype(np.uint8)
    print('running grabcut')
    #print('prior_mask.dtype = %r' % (prior_mask.dtype,))
    #print('bgr_img.dtype = %r' % (bgr_img.dtype,))
    with ut.Timer('grabcut'):
        post_mask = grabcut(bgr_img, prior_mask)
    if post_mask.dtype == np.uint8:
        post_mask = post_mask.astype(float) / 255.0
    seg_chip = mask_colored_img(bgr_img, post_mask, 'bgr')
    print('finished running grabcut')
    pt.imshow(post_mask * 255, pnum=(1, 2, 1))
    pt.imshow(seg_chip, pnum=(1, 2, 2))


def grabcut(bgr_img, prior_mask, binary=True):
    """
    Referencs:
        http://docs.opencv.org/trunk/doc/py_tutorials/py_imgproc/py_grabcut/py_grabcut.html
    """
    # Grab Cut Parameters
    (h, w) = bgr_img.shape[0:2]
    rect = (0, 0, w, h)
    num_iters = 5
    mode = cv2.GC_INIT_WITH_MASK
    bgd_model = np.zeros((1, 13 * 5), np.float64)
    fgd_model = np.zeros((1, 13 * 5), np.float64)
    # Grab Cut Execution
    post_mask = prior_mask.copy()
    cv2.grabCut(bgr_img, post_mask, rect, bgd_model, fgd_model, num_iters, mode=mode)
    if binary:
        is_forground = (post_mask == cv2.GC_FGD) + (post_mask == cv2.GC_PR_FGD)
        post_mask = np.where(is_forground, 255, 0).astype('uint8')
    else:
        label_colors = [       255,           170,            50,          0]
        label_values = [cv2.GC_FGD, cv2.GC_PR_FGD, cv2.GC_PR_BGD, cv2.GC_BGD]
        pos_list = [post_mask == value for value in label_values]
        for pos, color in zip(pos_list, label_colors):
            post_mask[pos] = color
    return post_mask


into_hsv_flags = {
    'bgr': cv2.COLOR_BGR2HSV,
    'rgb': cv2.COLOR_RGB2HSV,
}

from_hsv_flags = {
    'bgr': cv2.COLOR_HSV2BGR,
}


def mask_colored_img(img_rgb, mask, encoding='bgr'):
    if mask.dtype == np.uint8:
        mask = mask.astype(float) / 255.0
    into_hsv_flag = into_hsv_flags[encoding]
    from_hsv_flag = from_hsv_flags[encoding]
    # Mask out value component
    img_hsv = cv2.cvtColor(img_rgb, into_hsv_flag)
    img_hsv = np.array(img_hsv, dtype=float) / 255.0
    VAL_INDEX = 2
    img_hsv[:, :, VAL_INDEX] *= mask
    img_hsv = np.array(np.round(img_hsv * 255.0), dtype=np.uint8)
    masked_img_rgb = cv2.cvtColor(img_hsv, from_hsv_flag)
    return masked_img_rgb


# Open CV relevant values:
# grabcut_mode = cv2.GC_EVAL
# grabcut_mode = cv2.GC_INIT_WITH_RECT
# cv2.GC_BGD, cv2.GC_PR_BGD, cv2.GC_PR_FGD, cv2.GC_FGD
#@profile
def grabcut2(rgb_chip):
    (h, w) = rgb_chip.shape[0:2]
    _mask = np.zeros((h, w), dtype=np.uint8)  # Initialize: mask
    # Set inside to cv2.GC_PR_FGD (probably forground)
    _mask[ :, :] = cv2.GC_PR_FGD
    # Set border to cv2.GC_BGD (definitely background)
    _mask[ 0, :] = cv2.GC_BGD
    _mask[-1, :] = cv2.GC_BGD
    _mask[:,  0] = cv2.GC_BGD
    _mask[:, -1] = cv2.GC_BGD
    # Grab Cut Parameters
    rect = (0, 0, w, h)
    num_iters = 5
    mode = cv2.GC_INIT_WITH_MASK
    bgd_model = np.zeros((1, 13 * 5), np.float64)
    fgd_model = np.zeros((1, 13 * 5), np.float64)
    # Grab Cut Execution
    cv2.grabCut(rgb_chip, _mask, rect, bgd_model, fgd_model, num_iters, mode=mode)
    is_forground = (_mask == cv2.GC_FGD) + (_mask == cv2.GC_PR_FGD)
    chip_mask = np.where(is_forground, 255, 0).astype('uint8')
    # Crop
    chip_mask = clean_mask(chip_mask)
    chip_mask = np.array(chip_mask, float) / 255.0
    # Mask value component of HSV space
    seg_chip = mask_colored_img(rgb_chip, chip_mask, 'rgb')
    return seg_chip


def segment(img_fpath, bbox_, new_size=None):
    """ Runs grabcut """
    printDBG('[segm] segment(img_fpath=%r, bbox=%r)>' % (img_fpath, bbox_))
    num_iters = 5
    bgd_model = np.zeros((1, 13 * 5), np.float64)
    fgd_model = np.zeros((1, 13 * 5), np.float64)
    mode = cv2.GC_INIT_WITH_MASK
    # Initialize
    # !!! CV2 READS (H,W) !!!
    #  WH Unsafe
    img_resz, bbox_resz = resize_img_and_bbox(img_fpath, bbox_, new_size=new_size)
    # WH Unsafe
    (img_h, img_w) = img_resz.shape[:2]                       # Image Shape
    printDBG(' * img_resz.shape=%r' % ((img_h, img_w),))
    # WH Safe
    tlbr = ut.xywh_to_tlbr(bbox_resz, (img_w, img_h))  # Rectangle ANNOTATION
    (x1, y1, x2, y2) = tlbr
    rect = tuple(bbox_resz)                               # Initialize: rect
    printDBG(' * rect=%r' % (rect,))
    printDBG(' * tlbr=%r' % (tlbr,))
    # WH Unsafe
    _mask = np.zeros((img_h, img_w), dtype=np.uint8)  # Initialize: mask
    _mask[y1:y2, x1:x2] = cv2.GC_PR_FGD             # Set ANNOTATION to cv2.GC_PR_FGD
    # Grab Cut
    tt = ut.Timer(' * cv2.grabCut()', verbose=DEBUG_SEGM)
    cv2.grabCut(img_resz, _mask, rect, bgd_model, fgd_model, num_iters, mode=mode)
    tt.toc()
    img_mask = np.where((_mask == cv2.GC_FGD) + (_mask == cv2.GC_PR_FGD), 255, 0).astype('uint8')
    # Crop
    chip      = img_resz[y1:y2, x1:x2]
    chip_mask = img_mask[y1:y2, x1:x2]
    chip_mask = clean_mask(chip_mask)
    chip_mask = np.array(chip_mask, float) / 255.0
    # Mask the value of HSV
    chip_hsv = cv2.cvtColor(chip, cv2.COLOR_RGB2HSV)
    chip_hsv = np.array(chip_hsv, dtype=float) / 255.0
    chip_hsv[:, :, 2] *= chip_mask
    chip_hsv = np.array(np.round(chip_hsv * 255.0), dtype=np.uint8)
    seg_chip = cv2.cvtColor(chip_hsv, cv2.COLOR_HSV2RGB)
    return seg_chip, img_mask


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.segmentation
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



from PIL import Image
import utool as ut


def open_pil_image(image_fpath):
    pil_img = Image.open(image_fpath)
    return pil_img


def print_image_checks(img_fpath):
    hasimg = ut.checkpath(img_fpath, verbose=True)
    if hasimg:
        _tup = (img_fpath, ut.filesize_str(img_fpath))
        print('[io] Image %r (%s) exists. Is it corrupted?' % _tup)
    else:
        print('[io] Image %r does not exist ' % (img_fpath,))
    return hasimg
